@@ -27,13 +27,22 @@ module Hledger.Data.Transaction
 , transactionTransformPostings
 , transactionApplyValuation
 , transactionToCost
-, transactionAddInferredEquityPostings
-, transactionInferCostsFromEquity
+, transactionInferEquityPostings
+, transactionTagCostsAndEquityAndMaybeInferCosts
 , transactionApplyAliases
 , transactionMapPostings
 , transactionMapPostingAmounts
 , transactionAmounts
+, transactionCommodityStyles
+, transactionCommodityStylesWith
+, transactionNegate
 , partitionAndCheckConversionPostings
+, transactionAddTags
+, transactionAddHiddenAndMaybeVisibleTag
+  -- * helpers
+, TransactionBalancingPrecision(..)
+, payeeAndNoteFromDescription
+, payeeAndNoteFromDescription'
   -- nonzerobalanceerror
   -- * date operations
 , transactionDate2
@@ -46,7 +55,6 @@ module Hledger.Data.Transaction
 , showTransaction
 , showTransactionOneLineAmounts
 , showTransactionLineFirstPart
-, showTransactionBeancount
 , transactionFile
   -- * transaction errors
 , annotateErrorWithTransaction
@@ -74,15 +82,33 @@ import Hledger.Data.Amount
 import Hledger.Data.Valuation
 import Data.Decimal (normalizeDecimal, decimalPlaces)
 import Data.Functor ((<&>))
+import Data.Function ((&))
+import Data.List (union)
 
 
+-- | How to determine the precision used for checking that transactions are balanced. See #2402.
+data TransactionBalancingPrecision =
+    TBPOld
+    -- ^ Legacy behaviour, as in hledger <1.50, included to ease upgrades.
+    -- use precision inferred from the whole journal, overridable by commodity directive or -c.
+    -- Display precision is also transaction balancing precision; increasing it can break journal reading.
+    -- Some valid journals are rejected until commodity directives are added.
+    -- Small unbalanced remainders can be hidden, and in accounts that are never reconciled, can accumulate over time.
+  | TBPExact
+    -- ^ Simpler, more robust behaviour, as in Ledger: use precision inferred from the transaction.
+    -- Display precision and transaction balancing precision are independent; display precision never affects journal reading.
+    -- Valid journals from ledger or beancount are accepted without needing commodity directives.
+    -- Every imbalance in a transaction is visibly accounted for in that transaction's journal entry.
+
+  deriving (Bounded, Enum, Eq, Ord, Read, Show)
+
 instance HasAmounts Transaction where
   styleAmounts styles t = t{tpostings=styleAmounts styles $ tpostings t}
 
 nulltransaction :: Transaction
 nulltransaction = Transaction {
                     tindex=0,
-                    tsourcepos=nullsourcepos,
+                    tsourcepos=nullsourcepospair,
                     tdate=nulldate,
                     tdate2=Nothing,
                     tstatus=Unmarked,
@@ -176,33 +202,6 @@ showTransactionLineFirstPart t = T.concat [date, status, code]
            | otherwise            = ""
     code = if T.null (tcode t) then "" else wrap " (" ")" $ tcode t
 
--- | Like showTransaction, but generates Beancount journal format.
-showTransactionBeancount :: Transaction -> Text
-showTransactionBeancount t =
-  -- https://beancount.github.io/docs/beancount_language_syntax.html
-  -- similar to showTransactionHelper, but I haven't bothered with Builder
-     firstline <> nl
-  <> foldMap ((<> nl)) newlinecomments
-  <> foldMap ((<> nl)) (postingsAsLinesBeancount $ tpostings t)
-  <> nl
-  where
-    nl = "\n"
-    firstline = T.concat [date, status, payee, note, tags, samelinecomment]
-    date = showDate $ tdate t
-    status = if tstatus t == Pending then " !" else " *"
-    (payee,note) =
-      case payeeAndNoteFromDescription' $ tdescription t of
-        ("","") -> ("",      ""      )
-        ("",n ) -> (""     , wrapq n )
-        (p ,"") -> (wrapq p, wrapq "")
-        (p ,n ) -> (wrapq p, wrapq n )
-      where
-        wrapq = wrap " \"" "\"" . escapeDoubleQuotes . escapeBackslash
-    tags = T.concat $ map ((" #"<>).fst) $ ttags t
-    (samelinecomment, newlinecomments) =
-      case renderCommentLines (tcomment t) of []   -> ("",[])
-                                              c:cs -> (c,cs)
-
 hasRealPostings :: Transaction -> Bool
 hasRealPostings = not . null . realPostings
 
@@ -260,10 +259,11 @@ transactionApplyValuation priceoracle styles periodlast today v =
 transactionToCost :: ConversionOp -> Transaction -> Transaction
 transactionToCost cost t = t{tpostings = mapMaybe (postingToCost cost) $ tpostings t}
 
--- | Add inferred equity postings to a 'Transaction' using transaction prices.
-transactionAddInferredEquityPostings :: Bool -> AccountName -> Transaction -> Transaction
-transactionAddInferredEquityPostings verbosetags equityAcct t =
-    t{tpostings=concatMap (postingAddInferredEquityPostings verbosetags equityAcct) $ tpostings t}
+-- | For any costs in this 'Transaction' which don't have associated equity conversion postings,
+-- generate and add those.
+transactionInferEquityPostings :: Bool -> AccountName -> Transaction -> Transaction
+transactionInferEquityPostings verbosetags equityAcct t =
+  t{tpostings=concatMap (postingAddInferredEquityPostings verbosetags equityAcct) $ tpostings t}
 
 type IdxPosting = (Int, Posting)
 
@@ -273,17 +273,35 @@ type IdxPosting = (Int, Posting)
 
 label s = ((s <> ": ")++)
 
--- | Add costs inferred from equity postings in this transaction.
--- The name(s) of conversion equity accounts should be provided.
--- For every adjacent pair of conversion postings, it will first search the postings
--- with costs to see if any match. If so, it will tag these as matched.
--- If no postings with costs match, it will then search the postings without costs,
--- and will match the first such posting which matches one of the conversion amounts.
--- If it finds a match, it will add a cost and then tag it.
--- If the first argument is true, do a dry run instead: identify and tag
--- the costful and conversion postings, but don't add costs.
-transactionInferCostsFromEquity :: Bool -> [AccountName] -> Transaction -> Either String Transaction
-transactionInferCostsFromEquity dryrun conversionaccts t = first (annotateErrorWithTransaction t . T.unpack) $ do
+-- | Add tags to a transaction, discarding any for which it already has a value.
+-- Note this does not add tags to the transaction's comment.
+transactionAddTags :: Transaction -> [Tag] -> Transaction
+transactionAddTags t@Transaction{ttags} tags = t{ttags=ttags `union` tags}
+
+-- | Add the given hidden tag to a transaction; and with a true argument,
+-- also add the equivalent visible tag to the transaction's tags and comment fields.
+-- If the transaction already has these tags (with any value), do nothing.
+transactionAddHiddenAndMaybeVisibleTag :: Bool -> HiddenTag -> Transaction -> Transaction
+transactionAddHiddenAndMaybeVisibleTag verbosetags ht t@Transaction{tcomment=c, ttags} =
+  (t `transactionAddTags` ([ht] <> [ vt|verbosetags]))
+  {tcomment=if verbosetags && not hadtag then c `commentAddTagNextLine` vt else c}
+  where
+    vt@(vname,_) = toVisibleTag ht
+    hadtag = any ((== (T.toLower vname)) . T.toLower . fst) ttags  -- XXX should regex-quote vname
+
+-- | Find, associate, and tag the corresponding equity conversion postings and costful or potentially costful postings in this transaction.
+-- With a true addcosts argument, also generate and add any equivalent costs that are missing.
+-- The (previously detected) names of all equity conversion accounts should be provided.
+--
+-- For every pair of adjacent conversion postings, this first searches for a posting with equivalent cost (1).
+-- If no such posting is found, it then searches the costless postings, for one matching one of the conversion amounts (2).
+-- If either of these found a candidate posting, it is tagged with costPostingTagName.
+-- Then if in addcosts mode, if a costless posting was found, a cost equivalent to the conversion amounts is added to it.
+--
+-- The name reflects the complexity of this and its helpers; clarification is ongoing.
+--
+transactionTagCostsAndEquityAndMaybeInferCosts :: Bool -> Bool -> [AccountName] -> Transaction -> Either String Transaction
+transactionTagCostsAndEquityAndMaybeInferCosts verbosetags1 addcosts conversionaccts t = first (annotateErrorWithTransaction t . T.unpack) $ do
   -- number the postings
   let npostings = zip [0..] $ tpostings t
 
@@ -295,7 +313,7 @@ transactionInferCostsFromEquity dryrun conversionaccts t = first (annotateErrorW
   -- 1. each pair of conversion postings, and the corresponding postings which balance them, are tagged for easy identification
   -- 2. each pair of balancing postings which did't have an explicit cost, have had a cost calculated and added to one of them
   -- 3. if any ambiguous situation was detected, an informative error is raised
-  processposting <- transformIndexedPostingsF (addCostsToPostings dryrun) conversionPairs otherps
+  processposting <- transformIndexedPostingsF (tagAndMaybeAddCostsForEquityPostings verbosetags1 addcosts) conversionPairs otherps
 
   -- And if there was no error, use it to modify the transaction's postings.
   return t{tpostings = map (snd . processposting) npostings}
@@ -304,7 +322,7 @@ transactionInferCostsFromEquity dryrun conversionaccts t = first (annotateErrorW
 
     -- Generate the tricksy processposting function,
     -- which when applied to each posting in turn, rather magically has the effect of
-    -- applying addCostsToPostings to each pair of conversion postings in the transaction,
+    -- applying tagAndMaybeAddCostsForEquityPostings to each pair of conversion postings in the transaction,
     -- matching them with the other postings, tagging them and perhaps adding cost information to the other postings.
     -- General type:
     -- transformIndexedPostingsF :: (Monad m, Foldable t, Traversable t) =>
@@ -314,45 +332,44 @@ transactionInferCostsFromEquity dryrun conversionaccts t = first (annotateErrorW
     --   m (a1 -> a1)
     -- Concrete type:
     transformIndexedPostingsF ::
-      ((IdxPosting, IdxPosting) -> StateT ([IdxPosting],[IdxPosting]) (Either Text) (IdxPosting -> IdxPosting)) ->  -- state update function (addCostsToPostings with the bool applied)
+      ((IdxPosting, IdxPosting) -> StateT ([IdxPosting],[IdxPosting]) (Either Text) (IdxPosting -> IdxPosting)) ->  -- state update function (tagAndMaybeAddCostsForEquityPostings with the bool applied)
       [(IdxPosting, IdxPosting)] ->   -- initial state: the pairs of adjacent conversion postings in the transaction
       ([IdxPosting],[IdxPosting]) ->  -- initial state: the other postings in the transaction, separated into costful and costless
       (Either Text (IdxPosting -> IdxPosting))  -- returns an error message or a posting transform function
     transformIndexedPostingsF updatefn = evalStateT . fmap (appEndo . foldMap Endo) . traverse (updatefn)
 
     -- A tricksy state update helper for processposting/transformIndexedPostingsF.
-    -- Approximately: given a pair of conversion postings to match,
+    -- Approximately: given a pair of equity conversion postings to match,
     -- and lists of the remaining unmatched costful and costless other postings,
-    -- 1. find (and consume) two other postings which match the two conversion postings
-    -- 2. add identifying tags to the four postings
-    -- 3. add an explicit cost, if missing, to one of the matched other postings
-    -- 4. or if there is a problem, raise an informative error or do nothing as appropriate.
-    -- Or, if the first argument is true:
-    -- do a dry run instead: find and consume, add tags, but don't add costs
-    -- (and if there are no costful postings at all, do nothing).
-    addCostsToPostings :: Bool -> (IdxPosting, IdxPosting) -> StateT ([IdxPosting], [IdxPosting]) (Either Text) (IdxPosting -> IdxPosting)
-    addCostsToPostings dryrun' ((n1, cp1), (n2, cp2)) = StateT $ \(costps, otherps) -> do
+    -- 1. find (and consume) two other postings whose amounts/cost match the two conversion postings
+    -- 2. add hidden identifying tags to the conversion postings and the other posting which has (or could have) an equivalent cost
+    -- 3. if in add costs mode, and the potential equivalent-cost posting does not have that explicit cost, add it
+    -- 4. or if there is a problem, raise an informative error or do nothing, as appropriate.
+    -- Or if there are no costful postings at all, do nothing.
+    tagAndMaybeAddCostsForEquityPostings :: Bool -> Bool -> (IdxPosting, IdxPosting) -> StateT ([IdxPosting], [IdxPosting]) (Either Text) (IdxPosting -> IdxPosting)
+    tagAndMaybeAddCostsForEquityPostings verbosetags addcosts' ((n1, cp1), (n2, cp2)) = StateT $ \(costps, otherps) -> do
       -- Get the two conversion posting amounts, if possible
       ca1 <- conversionPostingAmountNoCost cp1
       ca2 <- conversionPostingAmountNoCost cp2
       let 
-        -- All costful postings which match the conversion posting pair
-        matchingCostPs =
+        -- All costful postings whose cost is equivalent to the conversion postings' amounts.
+        matchingCostfulPs =
           dbg7With (label "matched costful postings".show.length) $ 
           mapMaybe (mapM $ costfulPostingIfMatchesBothAmounts ca1 ca2) costps
 
-        -- All other single-commodity postings whose amount matches at least one of the conversion postings,
-        -- with an explicit cost added. Or in dry run mode, all other single-commodity postings.
-        matchingOtherPs =
+        -- In dry run mode: all other costless, single-commodity postings.
+        -- In add costs mode: all other costless, single-commodity postings whose amount matches at least one of the conversion postings,
+        -- with the equivalent cost added to one of them. (?)
+        matchingCostlessPs =
           dbg7With (label "matched costless postings".show.length) $
-          if dryrun'
-          then [(n,(p, a)) | (n,p) <- otherps, let Just a = postingSingleAmount p]
-          else mapMaybe (mapM $ addCostIfMatchesOneAmount ca1 ca2) otherps
+          if addcosts'
+          then mapMaybe (mapM $ addCostIfMatchesOneAmount ca1 ca2) otherps
+          else [(n,(p, a)) | (n,p) <- otherps, let Just a = postingSingleAmount p]
 
         -- A function that adds a cost and/or tag to a numbered posting if appropriate.
         postingAddCostAndOrTag np costp (n,p) =
-          (n, if | n == np            -> costp `postingAddTags` [("_price-matched","")]
-                 | n == n1 || n == n2 -> p     `postingAddTags` [("_conversion-matched","")]
+          (n, if | n == np            -> costp & postingAddHiddenAndMaybeVisibleTag verbosetags (costPostingTagName,"")        -- if it's the specified posting number, replace it with the costful posting, and tag it
+                 | n == n1 || n == n2 -> p     & postingAddHiddenAndMaybeVisibleTag verbosetags (conversionPostingTagName,"")  -- if it's one of the equity conversion postings, tag it
                  | otherwise          -> p)
 
       -- Annotate any errors with the conversion posting pair
@@ -362,20 +379,20 @@ transactionInferCostsFromEquity dryrun conversionaccts t = first (annotateErrorW
           -- delete it from the list of costful postings in the state, delete the
           -- first matching costless posting from the list of costless postings
           -- in the state, and return the transformation function with the new state.
-          | [(np, costp)] <- matchingCostPs
+          | [(np, costp)] <- matchingCostfulPs
           , Just newcostps <- deleteIdx np costps
-              -> Right (postingAddCostAndOrTag np costp, (if dryrun' then costps else newcostps, otherps))
+              -> Right (postingAddCostAndOrTag np costp, (if addcosts' then newcostps else costps, otherps))
 
           -- If no costful postings match the conversion postings, but some
           -- of the costless postings match, check that the first such posting has a
           -- different amount from all the others, and if so add a cost to it,
           -- then delete it from the list of costless postings in the state,
           -- and return the transformation function with the new state.
-          | [] <- matchingCostPs
-          , (np, (costp, amt)):nps <- matchingOtherPs
+          | [] <- matchingCostfulPs
+          , (np, (costp, amt)):nps <- matchingCostlessPs
           , not $ any (amountsMatch amt . snd . snd) nps
           , Just newotherps <- deleteIdx np otherps
-              -> Right (postingAddCostAndOrTag np costp, (costps, if dryrun' then otherps else newotherps))
+              -> Right (postingAddCostAndOrTag np costp, (costps, if addcosts' then newotherps else otherps))
 
           -- Otherwise, do nothing, leaving the transaction unchanged.
           -- We don't want to be over-zealous reporting problems here
@@ -387,13 +404,13 @@ transactionInferCostsFromEquity dryrun conversionaccts t = first (annotateErrorW
     -- with the matching amount which must be present in another non-conversion posting.
     costfulPostingIfMatchesBothAmounts :: Amount -> Amount -> Posting -> Maybe Posting
     costfulPostingIfMatchesBothAmounts a1 a2 costfulp = do
-        a@Amount{aprice=Just _} <- postingSingleAmount costfulp
+        a@Amount{acost=Just _} <- postingSingleAmount costfulp
         if
            | dbgamtmatch 1 a1 a (amountsMatch (-a1) a)  &&  dbgcostmatch 2 a2 a (amountsMatch a2 (amountCost a)) -> Just costfulp
            | dbgamtmatch 2 a2 a (amountsMatch (-a2) a)  &&  dbgcostmatch 1 a1 a (amountsMatch a1 (amountCost a)) -> Just costfulp
            | otherwise -> Nothing
            where
-            dbgamtmatch  n a b = dbg7 ("conversion posting "     <>show n<>" "<>showAmount a<>" balances amount "<>showAmountWithoutPrice b <>" of costful posting "<>showAmount b<>" at precision "<>dbgShowAmountPrecision a<>" ?")
+            dbgamtmatch  n a b = dbg7 ("conversion posting "     <>show n<>" "<>showAmount a<>" balances amount "<>showAmountWithoutCost b <>" of costful posting "<>showAmount b<>" at precision "<>dbgShowAmountPrecision a<>" ?")
             dbgcostmatch n a b = dbg7 ("and\nconversion posting "<>show n<>" "<>showAmount a<>" matches cost "   <>showAmount (amountCost b)<>" of costful posting "<>showAmount b<>" at precision "<>dbgShowAmountPrecision a<>" ?") 
 
     -- Add a cost to a posting if it matches (negative) one of the
@@ -401,7 +418,7 @@ transactionInferCostsFromEquity dryrun conversionaccts t = first (annotateErrorW
     addCostIfMatchesOneAmount :: Amount -> Amount -> Posting -> Maybe (Posting, Amount)
     addCostIfMatchesOneAmount a1 a2 p = do
         a <- postingSingleAmount p
-        let newp cost = p{pamount = mixedAmount a{aprice = Just $ TotalPrice cost}}
+        let newp cost = p{pamount = mixedAmount a{acost = Just $ TotalCost cost}}
         if
            | amountsMatch (-a1) a -> Just (newp a2, a2)
            | amountsMatch (-a2) a -> Just (newp a1, a1)
@@ -409,8 +426,8 @@ transactionInferCostsFromEquity dryrun conversionaccts t = first (annotateErrorW
 
     -- Get the single-commodity costless amount from a conversion posting, or raise an error.
     conversionPostingAmountNoCost p = case postingSingleAmount p of
-        Just a@Amount{aprice=Nothing} -> Right a
-        Just Amount{aprice=Just _} -> Left $ annotateWithPostings [p] "Conversion postings must not have a cost:"
+        Just a@Amount{acost=Nothing} -> Right a
+        Just Amount{acost=Just _} -> Left $ annotateWithPostings [p] "Conversion postings must not have a cost:"
         Nothing                    -> Left $ annotateWithPostings [p] "Conversion postings must have a single-commodity amount:"
 
     -- Do these amounts look the same when compared at the first's display precision ?
@@ -445,7 +462,7 @@ partitionAndCheckConversionPostings check conversionaccts =
   -- Left fold processes postings in parse order, so that eg inferred costs
   -- will be added to the first (top-most) posting, not the last one.
   foldlM select (([], ([], [])), Nothing)
-    -- The costless other postings are somehow reversed still; "second (second reverse" fixes that.
+    -- The costless other postings are somehow reversed still; "second (second reverse)" fixes that.
     <&> fmap (second (second reverse) . fst)
   where
     select ((cs, others@(ps, os)), Nothing) np@(_, p)
@@ -457,7 +474,7 @@ partitionAndCheckConversionPostings check conversionaccts =
       | check          = Left "Conversion postings must occur in adjacent pairs"
       | otherwise      = Right ((cs, (ps, np:os)), Nothing)
     isConversion p = paccount p `elem` conversionaccts
-    hasCost p = isJust $ aprice =<< postingSingleAmount p
+    hasCost p = isJust $ acost =<< postingSingleAmount p
 
 -- | Get a posting's amount if it is single-commodity.
 postingSingleAmount :: Posting -> Maybe Amount
@@ -481,10 +498,25 @@ transactionMapPostings f t@Transaction{tpostings=ps} = t{tpostings=map f ps}
 transactionMapPostingAmounts :: (MixedAmount -> MixedAmount) -> Transaction -> Transaction
 transactionMapPostingAmounts f  = transactionMapPostings (postingTransformAmount f)
 
--- | All posting amounts from this transactin, in order.
+-- | All posting amounts from this transaction, in order.
 transactionAmounts :: Transaction -> [MixedAmount]
 transactionAmounts = map pamount . tpostings
 
+-- | Get the canonical amount styles inferred from this transaction's amounts.
+transactionCommodityStyles :: Transaction -> M.Map CommoditySymbol AmountStyle
+transactionCommodityStyles =
+  either (const mempty) id .  -- ignore style problems, commodityStylesFromAmounts doesn't report them currently
+  commodityStylesFromAmounts . concatMap (amountsRaw . pamount) . tpostings
+
+-- | Like transactionCommodityStyles, but attach a particular rounding strategy to the styles,
+-- affecting how they will affect display precisions when applied.
+transactionCommodityStylesWith :: Rounding -> Transaction -> M.Map CommoditySymbol AmountStyle
+transactionCommodityStylesWith r = amountStylesSetRounding r . transactionCommodityStyles
+
+-- | Flip the sign of this transaction's posting amounts (and balance assertion amounts).
+transactionNegate :: Transaction -> Transaction
+transactionNegate = transactionMapPostings postingNegate
+
 -- | The file path from which this transaction was parsed.
 transactionFile :: Transaction -> FilePath
 transactionFile Transaction{tsourcepos} = sourceName $ fst tsourcepos
@@ -604,7 +636,7 @@ tests_Transaction =
                  Transaction
                    0
                    ""
-                   nullsourcepos
+                   nullsourcepospair
                    (fromGregorian 2007 01 28)
                    Nothing
                    Unmarked
@@ -628,7 +660,7 @@ tests_Transaction =
               Transaction
                 0
                 ""
-                nullsourcepos
+                nullsourcepospair
                 (fromGregorian 2007 01 28)
                 Nothing
                 Unmarked
@@ -651,7 +683,7 @@ tests_Transaction =
               Transaction
                 0
                 ""
-                nullsourcepos
+                nullsourcepospair
                 (fromGregorian 2007 01 28)
                 Nothing
                 Unmarked
@@ -667,7 +699,7 @@ tests_Transaction =
               Transaction
                 0
                 ""
-                nullsourcepos
+                nullsourcepospair
                 (fromGregorian 2010 01 01)
                 Nothing
                 Unmarked
